# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""The file has utilities to regress on trial intermediate measurements.

This contains utilities to fit regression models that predict the objective
at a particular future step of ACTIVE trials. We notably support LightGBM
models and necessary datastructures.
"""

import copy
from typing import (Any, Callable, Dict, Optional, Tuple, Union)

from absl import logging
import attrs
import lightgbm.sklearn as lightgbm
import numpy as np
from scipy.interpolate.fitpack2 import InterpolatedUnivariateSpline
import six
from six.moves import range
from sklearn import model_selection
from vizier import algorithms as vza
from vizier import pyvizier
from vizier.pyvizier import converters


@attrs.define
class TrialData:
  """Light weight trial data class to be used for training regression models."""

  id: int
  learning_rate: float
  final_objective: float
  steps: list[int]
  objective_values: list[float]

  @classmethod
  def from_trial(
      cls,
      trial: pyvizier.Trial,
      learning_rate_param_name: str,
      metric_name: str,
      converter: converters.TimedLabelsExtractor,
  ):
    """Preprocess the pyvizier trial into an instance of the class.

    Args:
      trial: pyvizier.Trial containing trial to process.
      learning_rate_param_name: name of learning rate param
      metric_name: name of optimization metric
      converter: vizier tool to convert trials to times sequences

    Returns:
      returned_trial: the trial in TrialData format
    """

    learning_rate = trial.parameters.get(
        learning_rate_param_name, pyvizier.ParameterValue(0.0)
    ).value

    timedlabels = converter.convert([trial])[0]
    steps, values = (
        np.asarray(timedlabels.times, np.int32).reshape(-1).tolist(),
        timedlabels.labels[metric_name].reshape(-1).tolist(),
    )

    if trial.final_measurement and (
        metric_name in trial.final_measurement.metrics
    ):
      final_value = converter.metric_converters[0].convert(
          [trial.final_measurement]
      )[0]
    else:
      final_value = values[-1] if values else 0.0

    return cls(
        id=trial.id,
        learning_rate=learning_rate,
        final_objective=final_value,
        steps=steps,
        objective_values=values,
    )

  def extrapolate_trial_objective_value(self, max_num_steps: int):
    """Extend the measurements of self to max_num_steps.

    Args:
      max_num_steps: target steps to extend the measurement.
    """
    last_step = self.steps[-1]
    if last_step >= max_num_steps:
      return

    last_objective = self.objective_values[-1]
    self.steps.append(max_num_steps)
    self.objective_values.append(last_objective)


def _generate_interpolation_fn_from_trial(
    steps: list[int], values: list[float]
) -> Callable[[int], float]:
  """Generates an interpolation function from a trial's measurement data.

  Since different trials have evaluations at different step numbers,
  we need to be able to interpolate the objective value between steps
  in order to compare trials and regress against trial data. This function
  converts a trial into a function suitable for this use.

  Args:
    steps:  list of integers indicating the x-axis of the input data points.
    values: list of floats indicating the y-axis of the input data points. steps
      and values list contains the same number of elements.

  Returns:
    interpolation function that takes input a number t and returns
    interpolated value of objective function for this trial at t steps.
  """
  return InterpolatedUnivariateSpline(steps, values, k=1)


def _sort_dedupe_measurements(
    steps: list[Union[int, float]], values: list[float]
) -> Tuple[list[Union[int, float]], list[float]]:
  """Sort and remove duplicates in the trial's measurements.

  Args:
    steps: a list of integer measurement steps for a given trial.
    values: a list of objective values corresponding to the steps for a given
      trial.

  Returns:
    steps: a list of integer measurement steps after dedupe.
    values: a list of objective values corresponding to the steps after dedupe.
  """
  if isinstance(steps[0], float):
    # Dedupe is skipped when steps are not integers.
    return steps, values
  step_obj_dict = {}
  updated_steps = []
  updated_values = []
  for index in range(len(steps)):
    step_obj_dict[steps[index]] = values[index]
  last_step = None
  for step, value in sorted(six.iteritems(step_obj_dict)):
    if last_step is None or step > last_step:
      updated_steps.append(step)
      updated_values.append(value)
      last_step = step
  return updated_steps, updated_values


class GBMAutoRegressor(object):
  """Train and predict trial measurements using auto-regressive GBM model."""

  def __init__(
      self,
      target_step: Union[int, float],
      min_points: int,
      learning_rate_param_name: str,
      metric_name: str,
      converter: converters.TimedLabelsExtractor,
      gbdt_param_grid: Optional[Dict[str, Any]] = None,
      cv: int = 2,
      random_state: Optional[int] = None,
  ):
    """Initialize model params.

    Args:
      target_step: the step to compute the prediction according to existing data
        points.
      min_points: number of lag points in the auto-regressive model
      learning_rate_param_name: name of learning rate param
      metric_name: name of optimization metric
      converter: vizier tool to convert trials to times sequences
      gbdt_param_grid: parameter grid for CV grid search for lightGBM model
      cv: k in k-fold cross validation
      random_state: random state for GBDT model
    """
    self._target_step = target_step
    self._min_points = min_points
    self._converter = converter
    self._learning_rate_param_name = learning_rate_param_name
    self._metric_name = metric_name
    self._gbdt_param_grid = gbdt_param_grid or {
        "max_depth": [2, 3, 5],
        "n_estimators": [50, 100],
    }
    self._cv = cv
    self._random_state = random_state
    self._model: lightgbm.LGBMRegressor = None  # place holder for trained model
    self._best_params: Dict[str, Any] = None  # place holder for best parameters

  @property
  def is_trained(self) -> bool:
    return self._model is not None

  def train(self, completed: vza.CompletedTrials) -> None:
    """Trains a GBDT combined models for auto-regression given completed trials.

    Args:
      completed: Sequence of completed trials.

    Returns:
      Nothing. Updated `_model` and `_best_params` members of the class.
    """
    completed_trials = []
    for trial in completed.trials:
      completed_trials.append(
          TrialData.from_trial(
              trial,
              learning_rate_param_name=self._learning_rate_param_name,
              metric_name=self._metric_name,
              converter=self._converter,
          )
      )
    if len(completed_trials) < self._min_points + 1:
      logging.info(
          "Not enough completed trials (only %d) to train GBDT model.",
          len(completed_trials),
      )
      return
    feature_matrix = []
    targets = []
    for trialc in completed_trials:
      # only consider trials with at least min_points + 1 steps as otherwise
      # the features constructed are mostly interpolated
      if len(trialc.steps) < self._min_points + 1:
        continue
      trialc.extrapolate_trial_objective_value(self._target_step)
      tc_steps, tc_values = _sort_dedupe_measurements(
          trialc.steps, trialc.objective_values
      )
      trial_inter_func = _generate_interpolation_fn_from_trial(
          tc_steps, tc_values
      )
      for i, step in enumerate(trialc.steps):
        if i < self._min_points - 1 or step >= self._target_step:
          continue
        features = self._create_features_from_trial(trialc, i)
        feature_matrix.append(features)
        targets.append(trial_inter_func(self._target_step))
    feature_matrix = np.array(feature_matrix)
    logging.info("Feature matrix shape: %s", feature_matrix.shape)
    if feature_matrix.shape[0] <= (self._min_points + 1) / (
        1.0 - 1.0 / self._cv
    ):
      logging.info(
          "Not enough rows in feature matrix. "
          "This can happen when there are not enough measurements in "
          "the completed trials."
      )
      return
    targets = np.array(targets)
    gbdt_cv = model_selection.GridSearchCV(
        lightgbm.LGBMRegressor(random_state=self._random_state),
        self._gbdt_param_grid,
        cv=self._cv,
    )
    gbdt_cv = gbdt_cv.fit(feature_matrix, targets)
    self._best_params = gbdt_cv.best_params_
    gbm = lightgbm.LGBMRegressor(
        **self._best_params, random_state=self._random_state
    )
    self._model = gbm.fit(feature_matrix, targets)

  def predict(self, trial: pyvizier.Trial) -> Union[float, None]:
    """Estimate objective values at target_steps using autoregression algorithm.

    Args:
      trial: current pyvizier trial.

    Returns:
      prediction: the predicted objective value at target_steps.
      It returns None when the trial has less than min_points steps.
    """
    trial_data = TrialData.from_trial(
        trial,
        learning_rate_param_name=self._learning_rate_param_name,
        metric_name=self._metric_name,
        converter=self._converter,
    )
    if not self.is_trained:
      raise ValueError("Prediction cannot be performed before model training.")
    # Not enough features for prediction
    if len(trial_data.steps) < self._min_points:
      return None
    features = self._create_features_from_trial(
        trial_data, len(trial_data.steps) - 1
    )
    features = np.asarray(features).reshape(1, -1)
    return self._model.predict(features)[0]

  def _create_features_from_trial(
      self,
      trial: TrialData,
      end_index: int,
  ) -> list[float]:
    """Create feature vector for auto-regressive model from a trial.

    Args:
      trial: sequence of measurements as a trial
      end_index: last index in steps to be included in the feature set

    Returns:
      list of features as 1D list of size `min_points`.
    """
    if self._min_points > end_index + 1:
      raise ValueError("Not enough data before end_index for creating features")
    if end_index >= len(trial.steps):
      raise ValueError("Not enough indices in trials.steps")
    features = [trial.learning_rate]
    for j in range(self._min_points):
      # Difference from target_step and value at the step into `features`.
      features.append(self._target_step - trial.steps[end_index - j])
      features.append(trial.objective_values[end_index - j])
    return features


@attrs.define
class HallucinationOptions(object):
  """Options for hallucinated suggestions policy.

  Attributes:
    autoregressive_order: order of the auto regressive predictor.
    use_steps: if true we use `steps` for prediction, otherwise we use
      `elapsed_seconds`.
    learning_rate_param_name: name of learning rate parameter.
    gbdt_param_grid: param grid for tuning the gbdt regressor.
    min_completed_trials: only start hallucinations after this many completed
      trials.
    min_steps: only hallucinate a stopped trial which has greater than equal to
      these many steps.
    max_steps: maximum number of steps in a trial.
    random_state: random seed.
    elapsed_seconds_gap: When using steps, the hallucinated measurements
      `elapsed_seconds` is set to be the current elapsed_seconds` + this
      constant.
  """

  autoregressive_order: int = attrs.field(default=5)
  learning_rate_param_name: str = attrs.field(default="learning_rate")
  use_steps: bool = attrs.field(default=True)
  gbdt_param_grid: Dict[str, Any] = attrs.field(
      default={"max_depth": [2, 3, 5], "n_estimators": [50, 100]}
  )
  min_completed_trials: int = attrs.field(default=5)
  min_steps: int = attrs.field(default=5)
  max_steps: Optional[int] = attrs.field(default=None)
  random_state: Optional[int] = attrs.field(default=None)
  elapsed_seconds_gap: Optional[int] = attrs.field(default=0)


class GBMTrialHallucinator:
  """Regression based early stopping hallucinations."""

  def __init__(
      self,
      study_config: pyvizier.ProblemStatement,
      options: HallucinationOptions,
      verbose: int = 0,
  ):
    """Initialization.

    Args:
      study_config: pyvizier study config.
      options: hallucination options
      verbose: verbosity level. If set to > 0, then we log trial predictions.
    """
    self._options = options
    self._study_config = study_config
    if not self._study_config.metric_information.is_single_objective:
      raise ValueError("This policy only supports one objective.")
    self._metric = self._study_config.metric_information.item()
    self._converter = converters.TimedLabelsExtractor(
        [
            converters.DefaultModelOutputConverter(
                self._metric, flip_sign_for_minimization_metrics=False
            ),
        ],
        timestamp="steps" if self._options.use_steps else "elapsed_secs",
        value_extraction="raw",
    )
    self._max_steps = self._options.max_steps
    self._min_steps = self._options.min_steps
    # Place holder for model.
    self._model: Optional[GBMAutoRegressor] = None
    self._verbose = verbose

  def train(self, train_trials: vza.CompletedTrials):
    """Returns the selected trials to be tried out next.

    Args:
      train_trials: a collection of completed trials to be used for training.
    """

    # Decide whether to update stopped trials or not based on
    # `update_all_stopped_trials` flag and if there are enough train trials.
    # Finally, training an autoregressive model if the conditions are met.
    has_enough_data = len(train_trials.trials) >= max(
        self._options.autoregressive_order + 1,
        self._options.min_completed_trials,
    )
    if not has_enough_data:
      logging.info("Not enough train trials.")
      return

    self._max_steps = self._max_steps or int(
        np.percentile([len(t.measurements) for t in train_trials.trials], 95)
    )

    # If min_steps was not set, take 1/10th of the max_num_steps as its
    # value.
    self._min_steps = self._min_steps or int(self._max_steps / 10)

    if not self._max_steps:
      logging.info("Not updating stopped trials as max steps could not be set.")
      return

    global_autoregressive_model = GBMAutoRegressor(
        target_step=self._max_steps,
        min_points=self._options.autoregressive_order,
        learning_rate_param_name=self._options.learning_rate_param_name,
        metric_name=self._metric.name,
        converter=self._converter,
        gbdt_param_grid=self._options.gbdt_param_grid,
        random_state=self._options.random_state,
    )
    global_autoregressive_model.train(train_trials)
    logging.info("Finished Training global Auto-regressive GBDT model.")

    # If `global_autoregressive_model` could not be trained, then stopped trials
    # are not updated before making suggestions.
    if not global_autoregressive_model.is_trained:
      logging.info(
          "Not updating stopped trials as global GBDT model is not trained."
      )
      return

    self._model = global_autoregressive_model

  def update_stopped_trials(
      self, stopped_trials: list[pyvizier.Trial]
  ) -> list[pyvizier.Trial]:
    """Add hallucinated final measurements to stopped trials.

    Args:
      stopped_trials: stopped trials.

    Returns:
      updated stopped trials with final measurement potentially added.
    """
    if not self._model:
      logging.info("Not updating stopped trials as model was not trained.")
      return stopped_trials
    updated_trials = []  # Collect all stopped trials that need to be updated.
    for pytrial in stopped_trials:
      if pytrial.infeasible:
        updated_trials.append(pytrial)
        continue
      auto_prediction = self._model.predict(pytrial)
      if not auto_prediction:
        updated_trials.append(pytrial)
        continue
      logging.log_if(
          logging.INFO,
          "Trial generated prediction %f",
          auto_prediction,
          self._verbose >= 1,
      )
      self._create_final_measurement(
          pytrial,
          auto_prediction=auto_prediction,
      )
      logging.info("Updated stopped trial : %s, using GBDT model.", pytrial.id)
      updated_trials.append(pytrial)
    return updated_trials

  def _create_final_measurement(
      self, pytrial: pyvizier.Trial, auto_prediction: float
  ):
    """Creates a final measurement for stopped trials."""
    if pytrial.final_measurement:
      logging.info(
          "A pending trial somehow has a final measurement and will"
          "remain unchanged"
          "(trial_id=%s)",
          pytrial.id,
      )
      return
    final_measurement = copy.deepcopy(pytrial.measurements[-1])
    final_measurement.metrics[self._metric.name] = pyvizier.Metric(
        value=auto_prediction
    )
    if self._options.use_steps:
      final_measurement.steps = self._max_steps
      # Increase the elapsed_secs for the final_measurement to
      # make sure new measurements won't have newer timestamps.
      final_measurement.elapsed_secs = (
          pytrial.measurements[-1].elapsed_secs
          + self._options.elapsed_seconds_gap
      )
    else:
      final_measurement.elapsed_secs = self._max_steps
      # Increase the steps for the final_measurement to ensure it's the last
      # step. Note that the gap does not effect vizier's suggestion policy
      # and therefore is a safe operation.
      if len(pytrial.measurements) > 1:
        checkpoint_gap = (
            pytrial.measurements[-1].steps - pytrial.measurements[-2].steps
        )
      else:
        checkpoint_gap = pytrial.measurements[-1].steps
      final_measurement.steps = (
          pytrial.measurements[-1].steps + checkpoint_gap * 2
      )
    pytrial.complete(final_measurement)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for pythia."""

import functools

import mock
import pyglove as pg
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.testing import test_runners
from vizier._src.pyglove import converters
from vizier._src.pyglove import pythia as pg_pythia
from vizier._src.pyglove import vizier_test_lib

from absl.testing import absltest

_RandomAlgorithm = vizier_test_lib.RandomAlgorithm


class TunerPolicyTest(absltest.TestCase):

  def test_stopping_policy(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('x', 0.0, 1.0)
    problem.metric_information.append(
        vz.MetricInformation(name='r', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
    )

    supporter = pythia.InRamPolicySupporter(problem)

    m = mock.create_autospec(pg.tuning.EarlyStoppingPolicy, instance=True)
    policy = pg_pythia.create_policy(
        supporter,
        problem,
        algorithm=_RandomAlgorithm(),
        early_stopping_policy=m,
    )

    supporter.AddTrials([
        vz.Trial(
            parameters={'x': 0.0},
            final_measurement=vz.Measurement({'r': 1.0}),
        ),
        vz.Trial(
            parameters={'x': 0.0},
            final_measurement=vz.Measurement({'r': 1.0}),
        ),
        vz.Trial(
            parameters={'x': 0.0},
            final_measurement=vz.Measurement({'r': 1.0}),
        ),
        vz.Trial(
            parameters={'x': 0.0},
            measurements=[vz.Measurement({'r': 1.0})],
        ),
    ])

    descriptor = supporter.study_descriptor()

    m.should_stop_early.return_value = False
    res = policy.early_stop(
        pythia.EarlyStopRequest(study_descriptor=descriptor, trial_ids=[4])
    )
    self.assertFalse(res.decisions[0].should_stop)

    m.should_stop_early.return_value = True
    res = policy.early_stop(
        pythia.EarlyStopRequest(study_descriptor=descriptor, trial_ids=[4])
    )
    self.assertTrue(res.decisions[0].should_stop)
    self.assertEqual(m.should_stop_early.call_count, 5)  # 3 + 1 + 1

  def test_random_algorithm_on_simple_search_space(self):
    """geno.Random wrapped into TunerPolicy should generate valid trials."""
    # Get a DNA spec.
    rewards = []

    def foo():
      r = pg.oneof([1, 2]) + pg.floatv(3., 4.)
      branch = pg.oneof(['negate', 'no-op'])
      if branch == 'negate':
        r = -r
      rewards.append(r)
      return r

    search_space = pg.hyper.DynamicEvaluationContext()
    with search_space.collect():
      foo()

    # Create a vizier converter.
    converter = converters.VizierConverter.from_dna_spec(
        search_space.dna_spec, ('',)
    )

    # Create pyglove algorithm.
    algorithm = _RandomAlgorithm()
    algorithm.setup(search_space.dna_spec)
    policy_factory = functools.partial(
        pg_pythia.TunerPolicy,
        converter=converter,
        algorithm=algorithm,
    )
    # Test the policy.
    tester = test_runners.RandomMetricsRunner(
        converter.problem,
        batch_size=5,
        validate_parameters=True,
    )
    tester.run_policy(policy_factory)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter base class for problem statement and evaluation.

Experimenters represent black-box optimization problems and/or users.
Each experimenter defines a ProblemStatement, representing the search space and
the metrics it returns in Evaluate (via CompletedTrials).

Pseudo-code for using Experimenters with Vizier Designers:

exp = ExperimenterSubClass(...)  # Possibly configure the experimenter.
problem_statement = exp.problem_statement()
designer = Designer(problem_statement)  # Configure the search algorithm
for i in range(10):
  suggestions = designer.suggest(count=2)
  exp.evaluate(suggestions) # Evaluate in-place, for maximum flexibility.
  designer.update(suggestions)
"""

import abc
from typing import Sequence

from vizier import pyvizier


class Experimenter(metaclass=abc.ABCMeta):
  """Abstract base class for Experimenters."""

  @abc.abstractmethod
  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    """Evaluates and mutates the Trials in-place.

    NOTE: The Experimenter is expected to mutate and/or complete the Trials as
    they wish, as to simulate users to maximum flexibility.

    Args:
      suggestions: Sequence of Trials to be evaluated.
    """
    pass

  @abc.abstractmethod
  def problem_statement(self) -> pyvizier.ProblemStatement:
    """The search configuration generated by this experimenter.

    The output should always be passed by value and not by reference, and thus
    should be generated inside this function or a deep copy of an existing
    problem statement.
    """
    pass



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Pythia Eagle Strategy Designer.

The designer implements a variation of Eagle Strategy without the Levy random
walk, aka Firefly Algorithm [1].

[1] Yang XS. (2009) Firefly Algorithms for Multimodal Optimization.
In: Stochastic Algorithms: Foundations and Applications (SAGA) 2009.
DOI: https://doi.org/10.1007/978-3-642-04944-6_14


Firefly Algorithm Summary
=========================
The Firefly algorithm is a genetic algorithm that operates by maintaining a
population of fireflies. Each firefly emits a light whose intensity corresponds
to the objective value being optimized. During each iteration, a firefly chases
after a brighter firefly in its vicinity, with the perceived brightness
diminishing as the distance increases. This feature fosters the formation of
multiple "clusters" of fireflies, as opposed to all fireflies converging to a
single point.

In addition to the inherent "attraction" forces guiding fireflies towards
brighter areas, we have incorporated a "repulsion" mechanism into the algorithm.
This means that fireflies not only move towards the brighter spots but also
actively move away from darker regions, enhancing the algorithm's exploration
capabilities.

We also incorporated support for various parameter types, including categorical,
discrete, and integer variables. We treat these parameter types uniquely when
computing distances, applying perturbations, and performing mutations on
fireflies, ensuring the algorithm's versatility in optimizing a wide range of
problem domains.

For more details about the Firefly algorithm, please refer to the linked paper.

OSS Vizier Implementation Summary
=================================
We maintain a dynamic pool of fireflies, with each firefly holding information
about the best trial it has produced. During the 'Suggest' phase, each firefly
takes its turn to propose a new trial. The best trial parameters are used to
calculate distances and generate fresh parameter suggestions by invoking the
'_mutate' and '_perturb' methods.

When the results of the suggested trials come in during the 'Update' phase,
we update/remove the corresponding firefly from the pool or create a new firefly
if necessary. To facilitate this association, during the 'Suggest' phase, each
newly sugggested trial stores in its metadata the identifier of its parent
firefly, denoted as 'parent_fly_id.' Over the course of the designer run, the
same firefly can be linked to multiple (child) trials that originated from it.

In our stateful implementation, across successive 'Suggest' calls, we
persistently maintain the firefly pool and update it using any new COMPLETED
trials that have not been seen before. We also ensure to not consecutively use
the same firefly for generating suggestions. Lastly we persist the maximum value
of firefly id created, to ensure that each firefly has its own unique id.
"""

import json
import time
from typing import Optional, Sequence

from absl import logging
import attr
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.core import abstractions
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy_utils
from vizier._src.algorithms.designers.eagle_strategy import serialization
from vizier.interfaces import serializable
from vizier.pyvizier import converters


EagleStrategyUtils = eagle_strategy_utils.EagleStrategyUtils
FireflyAlgorithmConfig = eagle_strategy_utils.FireflyAlgorithmConfig
Firefly = eagle_strategy_utils.Firefly
FireflyPool = eagle_strategy_utils.FireflyPool


class EagleStrategyDesigner(vza.PartiallySerializableDesigner):
  """The eagle strategy partially serializable designer."""

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      *,
      config: Optional[FireflyAlgorithmConfig] = None,
      seed: Optional[int] = None,
      initial_designer_factory: Optional[abstractions.DesignerFactory] = None,
  ):
    """Initializes the Eagle Strategy designer.

    Args:
      problem_statement: A problem description including the search space.
      config: The Firefly algorithm hyperparameters.
      seed: A seed to deterministically generate samples from random variables.
      initial_designer_factory: A partially serializable designer factory to
        generate the initial suggestions.

    Raises:
      Exception: if the problem statement includes condional search space,
        mutli-objectives or safety metrics.
    """
    # Problem statement validation.
    if problem_statement.search_space.is_conditional:
      raise ValueError(
          "Eagle Strategy designer doesn't support conditional parameters."
      )
    if not problem_statement.is_single_objective:
      raise ValueError(
          "Eagle Strategy designer doesn't support multi-objectives."
      )
    if problem_statement.is_safety_metric:
      raise ValueError(
          "Eagle Strategy designer doesn't support safety metrics."
      )

    if seed is None:
      # When a key is not provided it will be set based on the current time to
      # ensure non-repeated behavior.
      seed = int(time.time())
      logging.info(
          ('A seed was not provided to Eagle Strategy designer constructor. '
           'Setting the seed to %s'),
          str(seed),
      )
    self._scaler = converters.ProblemAndTrialsScaler(problem_statement)
    self._problem = self._scaler.problem_statement
    self._rng = np.random.default_rng(seed=seed)
    self._config = config or FireflyAlgorithmConfig()
    self._utils = EagleStrategyUtils(self._problem, self._config, self._rng)
    self._firefly_pool = FireflyPool(
        utils=self._utils, capacity=self._utils.compute_pool_capacity()
    )

    if initial_designer_factory is None:
      initial_designer_factory = quasi_random.QuasiRandomDesigner.from_problem
    self._initial_designer = initial_designer_factory(self._problem, seed=seed)

    logging.info(
        (
            'Eagle Strategy designer initialized. Pool capacity: %s. '
            'Eagle config:\n%s\nProblem statement:\n%s'
        ),
        self._utils.compute_pool_capacity(),
        json.dumps(attr.asdict(self._config), indent=2),
        self._problem,
    )

  def dump(self) -> vz.Metadata:
    """Dumps the current state of the algorithm.

    Returns:
      Metadata with the current pool serialized.
    """
    metadata = vz.Metadata()
    metadata.ns('eagle')['rng'] = serialization.serialize_rng(self._rng)
    metadata.ns('eagle')['firefly_pool'] = (
        serialization.partially_serialize_firefly_pool(self._firefly_pool))
    metadata.ns('eagle')['serialization_version'] = 'v1'
    metadata.ns('eagle')['dump_timestamp'] = str(time.time())
    metadata.ns('eagle').ns('random_designer').attach(
        self._initial_designer.dump()
    )
    return metadata

  def load(self, metadata: vz.Metadata) -> None:
    """Loads the current state of the algorithm run.

    The function is used as part of the `_restore_designer` method in
    PartiallySerializableDesignerPolicy to restore the state of the designer
    after it was already initialized with the problem statement.

    1. Set EagleStrategy, EagleStrategyUtils with the recovered random generator
    2. Recover the FireflyPool and populate it with the Fireflies.
    3. Recover the random designer state.

    Args:
      metadata: Metadata
    """
    if metadata.ns('eagle').get('serialization_version', default=None) is None:
      # First time the designer is called, so the namespace doesn't exist yet.
      logging.info('Eagle designer was called for the first time. No state was'
                   ' recovered.')
    else:
      try:
        self._rng = serialization.restore_rng(metadata.ns('eagle')['rng'])
      except Exception as e:
        raise serializable.FatalDecodeError(
            "Couldn't load random generator from metadata.") from e
      self._utils.rng = self._rng

      try:
        firefly_pool = metadata.ns('eagle')['firefly_pool']
        self._firefly_pool = serialization.restore_firefly_pool(
            self._utils, firefly_pool)
      except Exception as e:
        raise serializable.HarmlessDecodeError(
            "Couldn't load firefly pool from metadata.") from e

      try:
        self._initial_designer = quasi_random.QuasiRandomDesigner(
            self._problem.search_space
        )
        self._initial_designer.load(metadata.ns('eagle').ns('random_designer'))
      except Exception as e:
        raise serializable.HarmlessDecodeError(
            "Couldn't load random designer from metadata."
        ) from e

      logging.info(
          ('Eagle designer restored state from timestamp %s. Firefly pool'
           ' now contains %s fireflies.'),
          metadata.ns('eagle')['dump_timestamp'],
          self._firefly_pool.size,
      )

  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    """Suggests trials."""
    scaled_suggestions = [self._suggest_one() for _ in range(count)]
    #  Unscale suggestion parameters to the original search space.
    return self._scaler.unmap(scaled_suggestions)

  def _suggest_one(self) -> vz.TrialSuggestion:
    """Generates a single suggestion based on the current pool of flies.

    In order to generate a trial suggestion, we find the next fly that should be
    moved and create a copy of it. Then we mutate and perturb its trial
    parameters inplace and assign them to the suggested trial.

    Returns:
      The suggested trial with the parent fly Id in the metadata.
    """
    suggested_trial = vz.TrialSuggestion()
    if self._firefly_pool.size < self._firefly_pool.capacity:
      # Pool is underpopulated. Generate a random trial parameters.
      suggested_parameters = self._initial_designer.suggest()[0].parameters
      # Create a new parent fly id and assign it to the trial, this will be
      # used during Update to match the trial to its parent fly in the pool.
      parent_fly_id = self._firefly_pool.generate_new_fly_id()
    else:
      moving_fly = self._firefly_pool.get_next_moving_fly_copy()
      self._mutate_fly(moving_fly)
      self._perturb_fly(moving_fly)
      suggested_parameters = moving_fly.trial.parameters
      parent_fly_id = moving_fly.id_

    suggested_trial.parameters = suggested_parameters
    suggested_trial.metadata.ns('eagle')['parent_fly_id'] = str(parent_fly_id)
    return suggested_trial

  def _mutate_fly(self, moving_fly: Firefly) -> None:
    """Mutates fly's trial parameters inplace.

    Apply pulls from the rest of the pool's flies on `moving_fly` to mutate
    its trial's parameters.

    Args:
      moving_fly: the fire from the pool to mutate its trial parameters.
    """
    mutated_parameters = moving_fly.trial.parameters
    # Shuffle the ordering, so to apply the attracting and repelling forces
    # in random order every time. Creates a deep copy of the pool members.
    shuffled_flies = self._firefly_pool.get_shuffled_flies(self._rng)
    for other_fly in shuffled_flies:
      is_other_fly_better = self._utils.is_better_than(
          other_fly.trial, moving_fly.trial
      )
      # Compute 'other_fly' pull weights by parameter type. In the paper
      # this is called "attractivness" and is denoted by beta(r).
      pull_weights = self._utils.compute_pull_weight_by_type(
          other_fly.trial.parameters, mutated_parameters, is_other_fly_better
      )
      # Apply the pulls from 'other_fly' on the moving fly's parameters.
      for param_config in self._problem.search_space.parameters:
        pull_weight = pull_weights[param_config.type]
        if other_fly.trial.infeasible:
          pull_weight *= self._config.infeasible_force_factor
        # Accentuate 'other_fly' pull using 'exploration_rate'.
        if pull_weight > 0.5:
          explore_pull_weight = (
              self._config.explore_rate * pull_weight
              + (1 - self._config.explore_rate) * 1.0
          )
        else:
          explore_pull_weight = self._config.explore_rate * pull_weight
        # Update the parameters using 'other_fly' and 'explore_pull_rate'.
        mutated_parameters[param_config.name] = (
            self._utils.combine_two_parameters(
                param_config=param_config,
                param1=other_fly.trial.parameters,
                param2=mutated_parameters,
                param1_weight=explore_pull_weight,
            )
        )

  def _perturb_fly(self, moving_fly: Firefly) -> None:
    """Perturbs the fly's trial parameters inplace.

    Apply random perturbation to the fly's parameter values based on the
    moving_fly's 'perturbation' value.

    Args:
      moving_fly: the fire from the pool to mutate its trial parameters.
    """
    suggested_parameters = moving_fly.trial.parameters
    perturbations = self._utils.create_perturbations(moving_fly.perturbation)
    for i, param_config in enumerate(self._problem.search_space.parameters):
      perturbed_value = self._utils.perturb_parameter(
          param_config,
          suggested_parameters[param_config.name].value,
          perturbations[i],
      )
      suggested_parameters[param_config.name] = perturbed_value

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the pool.

    Iterate over new completed trials and update the firefly pool. Every trial
    that was suggested from Eagle Strategy has a metadata data containing the
    parent fly id. For trials that were added to the study externally we assign
    a new parent fly id.

    Trials passed to 'update' are in the unscaled/original search space, and
    will be converted to the scaled search space, so that all other methods
    in the designer deal with scaled trial values.

    Arguments:
      completed: Trials in the original search space.
      all_active:
    """
    del all_active
    trials = self._scaler.map(completed.trials)
    for trial in trials:
      # Replaces trial metric name with a canonical metric name, which makes the
      # serialization and deserialization simpler.
      trial = self._utils.standardize_trial_metric_name(trial)
      if not trial.metadata.ns('eagle').get('parent_fly_id'):
        # Trial was not generated from Eagle Strategy. Set a new parent fly id.
        trial.metadata.ns('eagle')['parent_fly_id'] = str(
            self._firefly_pool.generate_new_fly_id()
        )
      self._update_one(trial)

  def _update_one(self, trial: vz.Trial) -> None:
    """Update the pool using a single trial."""
    if trial.infeasible and self._config.infeasible_force_factor > 0:
      # Add the infeasible firefly to the pool.
      infeasible_firefly_id = self._firefly_pool.generate_new_fly_id()
      self._firefly_pool.create_or_update_fly(trial, infeasible_firefly_id)

    parent_fly_id = int(trial.metadata.ns('eagle').get('parent_fly_id'))
    parent_fly = self._firefly_pool.find_parent_fly(parent_fly_id)
    if parent_fly is None:
      if trial.infeasible:
        # Ignore infeasible trials without parent fly.
        pass
      elif self._firefly_pool.size < self._firefly_pool.capacity:
        # Pool is below capacity. Create a new firefly or update existing one.
        self._firefly_pool.create_or_update_fly(trial, parent_fly_id)
        return
      else:
        # Pool is at capacity. Try assigning a parent to utilize the trial info.
        parent_fly = self._assign_closest_parent(trial)

    if parent_fly is None:
      # Parent fly wasn't established. No need to continue.
      return

    elif not trial.infeasible and self._utils.is_better_than(
        trial, parent_fly.trial):
      # There's improvement. Update the parent with the new trial.
      parent_fly.trial = trial
      parent_fly.generation += 1
    else:
      # There's no improvement. Penalize the parent by decreasing its
      # exploration capability and potenitally remove it from the pool.
      self._penalize_parent_fly(parent_fly, trial)

  def _assign_closest_parent(self, trial: vz.Trial) -> Optional[Firefly]:
    """Finds the closest parent fly and checks that the trial improves on it.

    Note that the trial's `parent_fly_id` won't exist in the pool when:

    1. The trial was randomly generated during 'Suggest', and therefore is not
    associated with any existing fly in the pool.

    2. The fly was removed from the pool but not all associated trials were
    processed yet.

    3. The trial was added to the study independently of Eagle Strategy
    suggestions.

    Args:
      trial:

    Returns:
      None or a fly from the pool that is closest to the trial.
    """
    closest_parent_fly = self._firefly_pool.find_closest_parent(trial)
    if self._utils.is_better_than(trial, closest_parent_fly.trial):
      # Only returns the closest fly if there's improvement. Otherwise, we don't
      # return it to not count it as a failure, as the closest parent is not
      # reponsible for it.
      return closest_parent_fly

  def _penalize_parent_fly(self, parent_fly: Firefly, trial: vz.Trial) -> None:
    """Penalizes a parent fly.

    The method is called on a fly after its generated trial didn't improve the
    objective function. The fly is penalized by decreasing its exploration
    capability and potentially removing it from the pool.

    Args:
      parent_fly: The parent fly with to be penalized.
      trial: The generated trial from the parent fly that didn't imporove.
    """
    if trial.parameters == parent_fly.trial.parameters:
      # If the new trial is identical to the parent trial, it means that the
      # fly is stuck, and so we increase its perturbation.
      parent_fly.perturbation = min(
          parent_fly.perturbation * 10, self._config.max_perturbation
      )
    else:
      # Otherwise, penalize the parent by decreasing its perturbation factor.
      parent_fly.perturbation *= self._config.penalize_factor

    if parent_fly.perturbation < self._config.perturbation_lower_bound:
      # If the perturbation factor is too low we attempt to eliminate the
      # unsuccessful parent fly from the pool.
      if self._firefly_pool.size == self._firefly_pool.capacity:
        # Only remove if the pool is at capacity. This is critical in studies
        # with few feasible/safe trials to retain the feasible trials.
        if not self._firefly_pool.is_best_fly(parent_fly):
          # Check that the fly is not the best one we have thus far.
          self._firefly_pool.remove_fly(parent_fly)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for spatio_temporal."""

import numpy as np
from vizier import pyvizier
from vizier.pyvizier.converters import core
from vizier.pyvizier.converters import spatio_temporal as st

from absl.testing import absltest

_metric_converters = [
    core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='y1', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)),
    core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='y2', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
]

_trials = [
    pyvizier.Trial(
        id=1,
        parameters={'x1': pyvizier.ParameterValue(1)},
        measurements=[
            pyvizier.Measurement(
                steps=1, elapsed_secs=10, metrics={
                    'y1': 1,
                    'y2': -1
                }),
            pyvizier.Measurement(
                steps=2, elapsed_secs=20, metrics={
                    'y1': 3,
                    'y2': -3
                }),
            pyvizier.Measurement(
                steps=3, elapsed_secs=30, metrics={
                    'y1': 2,
                    'y2': -2
                })
        ]),
    pyvizier.Trial(
        id=2,
        parameters={'x1': pyvizier.ParameterValue(2)},
        measurements=[
            pyvizier.Measurement(
                steps=1, elapsed_secs=10, metrics={
                    'y1': -4,
                    'y2': 4
                }),
            pyvizier.Measurement(
                steps=5, elapsed_secs=50, metrics={
                    'y1': -6,
                    'y2': 6
                }),
            pyvizier.Measurement(
                steps=6, elapsed_secs=60, metrics={
                    'y1': -5,
                    'y2': 5
                })
        ])
]


class TimedLabelsExtractorTest(absltest.TestCase):

  def test_steps(self):
    converter = st.TimedLabelsExtractor(
        _metric_converters, 'steps', value_extraction='raw')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].times,
        np.asarray([1, 2, 3], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].times,
        np.asarray([1, 5, 6], dtype=np.float32)[:, np.newaxis])

  def test_elapsed_secs(self):
    converter = st.TimedLabelsExtractor(
        _metric_converters, 'elapsed_secs', value_extraction='raw')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].times,
        np.asarray([10, 20, 30], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].times,
        np.asarray([10, 50, 60], dtype=np.float32)[:, np.newaxis])

  def test_index(self):
    converter = st.TimedLabelsExtractor(
        _metric_converters, 'index', value_extraction='raw')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].times,
        np.asarray([0, 1, 2], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].times,
        np.asarray([0, 1, 2], dtype=np.float32)[:, np.newaxis])

  def test_labels_raw(self):
    converter = st.TimedLabelsExtractor(
        _metric_converters, 'elapsed_secs', value_extraction='raw')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].labels['y1'],
        np.asarray([1, 3, 2], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[0].labels['y2'],
        np.asarray([-1, -3, -2], dtype=np.float32)[:, np.newaxis])

  def test_labels_cummax(self):
    converter = st.TimedLabelsExtractor(
        _metric_converters, 'elapsed_secs', value_extraction='cummax')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].labels['y1'],
        np.asarray([1, 3, 3], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[0].labels['y2'],
        np.asarray([-1, -3, -3], dtype=np.float32)[:, np.newaxis])

  def test_labels_strict_cummax_firstonly(self):
    converter = st.TimedLabelsExtractor([_metric_converters[0]],
                                        'elapsed_secs',
                                        value_extraction='cummax_firstonly')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].times,
        np.asarray([10, 20, 30], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[0].labels['y1'],
        np.asarray([1, 3, 3], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].times,
        np.asarray([10, 60], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].labels['y1'],
        np.asarray([-4, -4], dtype=np.float32)[:, np.newaxis])

  def test_labels_cummax_lastonly(self):
    converter = st.TimedLabelsExtractor([_metric_converters[0]],
                                        'elapsed_secs',
                                        value_extraction='cummax_lastonly')
    timed_labels = converter.convert(_trials)
    np.testing.assert_almost_equal(
        timed_labels[0].times,
        np.asarray([10, 30], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[0].labels['y1'],
        np.asarray([1, 3], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].times,
        np.asarray([60], dtype=np.float32)[:, np.newaxis])
    np.testing.assert_almost_equal(
        timed_labels[1].labels['y1'],
        np.asarray([-4], dtype=np.float32)[:, np.newaxis])

  def test_extract_all_timestamps(self):
    converter = st.TimedLabelsExtractor(
        _metric_converters, 'steps', value_extraction='cummax')

    all_ts = converter.extract_all_timestamps(_trials)
    np.testing.assert_almost_equal(
        all_ts, np.asarray([1, 2, 3, 5, 6], dtype=np.float32))


class SparseSpatioTemporalConverterTest(absltest.TestCase):

  def test_all(self):
    extractor = st.TimedLabelsExtractor(
        _metric_converters, 'steps', value_extraction='raw')
    converter = st.SparseSpatioTemporalConverter([], extractor)
    features, labels = converter.to_xy(_trials)

    np.testing.assert_equal(features, {
        'steps':
            np.array([[1.], [2.], [3.], [1.], [5.], [6.]], dtype=np.float32)
    })
    np.testing.assert_equal(
        labels, {
            'y1': [[1.], [3.], [2.], [-4.], [-6.], [-5.]],
            'y2': [[-1.], [-3.], [-2.], [4.], [6.], [5.]],
        })

    self.assertEqual(converter.features_shape, {'steps': (None, 1)})
    self.assertEqual(converter.labels_shape, {'y1': (None, 1), 'y2': (None, 1)})
    self.assertEqual(
        converter.output_specs, {
            'steps':
                core.NumpyArraySpec.from_parameter_config(
                    pyvizier.ParameterConfig.factory(
                        name='steps', bounds=(0.0, np.finfo(float).max)),
                    core.NumpyArraySpecType.default_factory)
        })


class DenseSpatioTemporalConverterTest(absltest.TestCase):

  def test_all(self):
    extractor = st.TimedLabelsExtractor(
        _metric_converters, 'steps', value_extraction='raw')
    converter = st.DenseSpatioTemporalConverter([],
                                                extractor,
                                                temporal_index_points=np.array(
                                                    [1., 2., 3., 5., 6.]))
    features, labels = converter.to_xy(_trials)

    np.testing.assert_equal(features, {})
    self.assertEqual(converter.features_shape, {})

    np.testing.assert_equal(
        labels, {
            'y1':
                np.array([[1., 3., 2., np.nan, np.nan],
                          [-4., np.nan, np.nan, -6., -5.]]),
            'y2':
                np.array([[-1., -3., -2., np.nan, np.nan],
                          [4., np.nan, np.nan, 6., 5.]])
        },
        err_msg=f'{labels}')

    self.assertEqual(converter.labels_shape, {'y1': (None, 5), 'y2': (None, 5)})

  def test_xty(self):
    extractor = st.TimedLabelsExtractor(
        _metric_converters, 'steps', value_extraction='raw')
    parameter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory(name='x1', bounds=(0, 5)))
    converter = st.DenseSpatioTemporalConverter([parameter],
                                                extractor,
                                                temporal_index_points=np.array(
                                                    [1., 5., 6.]))
    features, temporal_index_points, labels = converter.to_xty(_trials, 'infer')

    np.testing.assert_equal(features, {'x1': [[1], [2]]})

    np.testing.assert_equal(temporal_index_points, [
        1.,
        2.,
        3.,
        5.,
        6.,
    ])

    np.testing.assert_equal(
        labels, {
            'y1':
                np.array([[1., 3., 2., np.nan, np.nan],
                          [-4., np.nan, np.nan, -6., -5.]]),
            'y2':
                np.array([[-1., -3., -2., np.nan, np.nan],
                          [4., np.nan, np.nan, 6., 5.]])
        },
        err_msg=f'{labels}')


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for eagle_optimizer."""

import logging
from typing import Optional

import jax
from jax import numpy as jnp
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import eagle_strategy
from vizier._src.algorithms.optimizers import random_vectorized_optimizer as rvo
from vizier._src.algorithms.testing import comparator_runner
from vizier._src.jax import types
from vizier.pyvizier import converters

from absl.testing import absltest
from absl.testing import parameterized


def create_continuous_problem(
    n_features: int,
    problem: Optional[vz.ProblemStatement] = None) -> vz.ProblemStatement:
  if not problem:
    problem = vz.ProblemStatement()
  root = problem.search_space.select_root()
  for i in range(n_features):
    root.add_float_param('x%d' % i, -5.0, 5.0)
  return problem


def create_categorical_problem(
    n_features: int,
    categorical_dim: int = 6,
    problem: Optional[vz.ProblemStatement] = None) -> vz.ProblemStatement:
  if not problem:
    problem = vz.ProblemStatement()
  root = problem.search_space.select_root()
  for i in range(n_features):
    root.add_categorical_param(
        'c%d' % i, feasible_values=[str(i) for i in range(categorical_dim)])
  return problem


def create_mix_problem(n_features: int,
                       categorical_dim: int = 8) -> vz.ProblemStatement:
  problem = create_continuous_problem(n_features // 2)
  return create_categorical_problem(n_features // 2, categorical_dim, problem)


# TODO: Change to bbob functions when they can support batching.
def sphere(x: types.ModelInput) -> jax.Array:
  return -(
      jnp.sum(jnp.square(x.continuous.padded_array), axis=-1)
      + 0.1 * jnp.sum(jnp.square(x.categorical.padded_array), axis=-1)
  )


def _rastrigin_d10_part(x: types.Array) -> jax.Array:
  return 10 * jnp.sum(jnp.cos(2 * np.pi * x), axis=-1) - jnp.sum(
      jnp.square(x), axis=-1
  )


def rastrigin_d10(x: types.ModelInput) -> jax.Array:
  return _rastrigin_d10_part(x.continuous.padded_array) + _rastrigin_d10_part(
      0.01 * x.categorical.padded_array
  )


class EagleOptimizerConvegenceTest(parameterized.TestCase):
  """Test optimizing an acquisition functions using vectorized Eagle Strategy.
  """

  @absltest.skip("Test takes too long externally.")
  @parameterized.product(
      create_problem_fn=[
          create_continuous_problem,
          create_categorical_problem,
          create_mix_problem,
      ],
      n_features=[10, 20],
      score_fn=[sphere, rastrigin_d10],
  )
  def test_converges(self, create_problem_fn, n_features, score_fn):
    logging.info('Starting a new convergence test (n_features: %s)', n_features)
    evaluations = 20_000
    problem = create_problem_fn(n_features)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    eagle_strategy_factory = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=eagle_strategy.EagleStrategyConfig())
    shifted_score_fn = score_fn
    random_strategy_factory = rvo.random_strategy_factory
    # Run simple regret convergence test.
    comparator_runner.SimpleRegretComparisonTester(
        baseline_num_trials=2 * evaluations,
        candidate_num_trials=evaluations,
        baseline_suggestion_batch_size=5,
        candidate_suggestion_batch_size=5,
        baseline_num_repeats=5,
        candidate_num_repeats=3,
        alpha=0.05,
        goal=vz.ObjectiveMetricGoal.MAXIMIZE,
    ).assert_optimizer_better_simple_regret(
        converter=converter,
        score_fn=shifted_score_fn,
        baseline_strategy_factory=random_strategy_factory,
        candidate_strategy_factory=eagle_strategy_factory,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Makes client calls via gRPC to an existing Vizier Service Server.

This client can be used interchangeably with the Cloud Vizier client.
"""

import datetime
import functools
import time
from typing import Any, Dict, List, Mapping, Optional, Union

from absl import logging
import attr
import grpc

from vizier._src.service import constants
from vizier._src.service import resources
from vizier._src.service import stubs_util
from vizier._src.service import study_pb2
from vizier._src.service import types
from vizier._src.service import vizier_service_pb2
from vizier._src.service import vizier_service_pb2_grpc
from vizier.service import pyvizier
from vizier.utils import attrs_utils

from google.longrunning import operations_pb2
from google.protobuf import duration_pb2
from google.protobuf import json_format


@attr.define
class _EnvironmentVariables:
  """Global environment variables.

  Attributes:
    server_endpoint: Endpoint to the Vizier server.
    servicer_kwargs:
    new_suggestion_polling_secs: The period to wait between polling for the
      status of long-running SuggestOperations. Vizier may increase this period
      if multiple polls are needed. (You may use zero for interactive demos, but
      it is only appropriate for very small Studies.)
  """

  server_endpoint: str = attr.field(
      default=constants.NO_ENDPOINT, validator=attr.validators.instance_of(str)
  )
  servicer_kwargs: Dict[str, Any] = attr.field(factory=dict)

  # TODO: Add an e2e test for this.
  new_suggestion_polling_secs: float = attr.field(default=1.0)

  def servicer_use_sql_ram(self) -> None:
    """Should be used in tests to avoid filepath issues."""
    self.servicer_kwargs['database_url'] = constants.SQL_MEMORY_URL


environment_variables = _EnvironmentVariables()


@functools.lru_cache(maxsize=None)
def _create_local_vizier_servicer() -> (
    vizier_service_pb2_grpc.VizierServiceServicer
):
  from vizier._src.service import vizier_service  # pylint:disable=g-import-not-at-top

  return vizier_service.VizierServicer(**environment_variables.servicer_kwargs)


def create_vizier_servicer_or_stub() -> types.VizierService:
  endpoint = environment_variables.server_endpoint
  if endpoint == constants.NO_ENDPOINT:
    logging.info('No endpoint given; using cached local VizierServicer.')
    logging.warning('Python 3.8+ is required in this case.')
    return _create_local_vizier_servicer()
  return stubs_util.create_vizier_server_stub(endpoint)


@attr.frozen(init=True)
class VizierClient:
  """Client for communicating with the Vizier Service via GRPC.

  It can be initialized directly with a VizierService, or
  created from endpoint. See also `create_server_stub`.
  """

  _study_resource_name: str = attr.field(
      validator=attr.validators.instance_of(str)
  )
  _client_id: str = attr.field(
      validator=[attr.validators.instance_of(str), attrs_utils.assert_not_empty]
  )
  _service: types.VizierService = attr.field(
      repr=False, factory=create_vizier_servicer_or_stub
  )

  @property
  def _study_resource(self) -> resources.StudyResource:
    return resources.StudyResource.from_name(self._study_resource_name)

  @property
  def _owner_id(self) -> str:
    return self._study_resource.owner_id

  @property
  def _study_id(self) -> str:
    return self._study_resource.study_id

  @property
  def study_resource_name(self) -> str:
    return self._study_resource_name

  def get_suggestions(
      self, suggestion_count: int, *, client_id_override: Optional[str] = None
  ) -> List[pyvizier.Trial]:
    """Gets a list of suggested Trials.

    Args:
        suggestion_count: The number of suggestions to request.
        client_id_override: If set, overrides self._client_id for this call.

    Returns:
      A list of PyVizier Trials. This may be an empty list if:
      1. A finite search space has been exhausted.
      2. If max_num_trials has been reached.
      3. Or if there are no longer any trials that match a supplied Context.

    Raises:
        RuntimeError: Indicates that a suggestion was requested
            from an inactive study. Note that this is NOT raised when a
            finite Study runs out of suggestions. In such a case, an empty
            list is returned.
    """
    if client_id_override is not None:
      client_id = client_id_override
    else:
      client_id = self._client_id
    request = vizier_service_pb2.SuggestTrialsRequest(
        parent=resources.StudyResource(self._owner_id, self._study_id).name,
        suggestion_count=suggestion_count,
        client_id=client_id,
    )
    try:
      operation = self._service.SuggestTrials(request)
    except grpc.RpcError as rpc_error:
      # If ImmutableStudyError occurs, we simply return empty suggestion list.
      # Otherwise, halt the client and raise error.
      if rpc_error.code() == grpc.StatusCode.FAILED_PRECONDITION:  # pytype:disable=attribute-error
        return []
      raise rpc_error

    num_attempts = 0
    while not operation.done:
      sleep_time = PollingDelay(
          num_attempts, environment_variables.new_suggestion_polling_secs
      )
      num_attempts += 1
      logging.info(
          'Waiting for operation with name %s to be done', operation.name
      )
      time.sleep(sleep_time.total_seconds())

      operation = self._service.GetOperation(
          operations_pb2.GetOperationRequest(name=operation.name)
      )

    if operation.HasField('error'):
      error_message = 'SuggestOperation {} failed with message: {}'.format(
          operation.name, operation.error
      )
      logging.error(error_message)
      raise RuntimeError(error_message)
    # TODO: Replace with any.Unpack().
    trials = vizier_service_pb2.SuggestTrialsResponse.FromString(
        operation.response.value
    ).trials
    return pyvizier.TrialConverter.from_protos(trials)

  def report_intermediate_objective_value(
      self,
      step: int,
      elapsed_secs: float,
      metric_list: List[Mapping[str, Union[int, float]]],
      trial_id: int,
  ) -> pyvizier.Trial:
    """Sends intermediate objective value for the trial identified by trial_id."""
    new_metric_list = []
    for metric in metric_list:
      for metric_name in metric:
        metric_pb2 = study_pb2.Measurement.Metric(
            metric_id=metric_name, value=metric[metric_name]
        )
        new_metric_list.append(metric_pb2)

    integer_seconds = int(elapsed_secs)
    nano_seconds = int((elapsed_secs - integer_seconds) * 1e9)
    measurement = study_pb2.Measurement(
        elapsed_duration=duration_pb2.Duration(
            seconds=integer_seconds, nanos=nano_seconds
        ),
        step_count=step,
        metrics=new_metric_list,
    )
    request = vizier_service_pb2.AddTrialMeasurementRequest(
        trial_name=resources.TrialResource(
            self._owner_id, self._study_id, trial_id
        ).name,
        measurement=measurement,
    )
    trial = self._service.AddTrialMeasurement(request)
    return pyvizier.TrialConverter.from_proto(trial)

  def should_trial_stop(self, trial_id: int) -> bool:
    request = vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
        trial_name=resources.TrialResource(
            self._owner_id, self._study_id, trial_id
        ).name
    )
    early_stopping_response = self._service.CheckTrialEarlyStoppingState(
        request
    )
    return early_stopping_response.should_stop

  def stop_trial(self, trial_id: int) -> None:
    request = vizier_service_pb2.StopTrialRequest(
        name=resources.TrialResource(
            self._owner_id, self._study_id, trial_id
        ).name
    )
    self._service.StopTrial(request)
    logging.info('Trial with id %s stopped.', trial_id)

  def complete_trial(
      self,
      trial_id: int,
      final_measurement: Optional[pyvizier.Measurement] = None,
      infeasibility_reason: Optional[str] = None,
  ) -> pyvizier.Trial:
    """Completes the trial, which is infeasible if given a infeasibility_reason."""
    request = vizier_service_pb2.CompleteTrialRequest(
        name=resources.TrialResource(
            self._owner_id, self._study_id, trial_id
        ).name,
        trial_infeasible=infeasibility_reason is not None,
        infeasible_reason=infeasibility_reason,
    )

    if final_measurement is not None:
      # Final measurement can still be included even for infeasible trials, for
      # other metrics, or a subset of objective + safety metrics.
      request.final_measurement.CopyFrom(
          pyvizier.MeasurementConverter.to_proto(final_measurement)
      )

    trial = self._service.CompleteTrial(request)
    return pyvizier.TrialConverter.from_proto(trial)

  def get_trial(self, trial_id: int) -> pyvizier.Trial:
    """Return the Optimizer trial for the given trial_id."""
    request = vizier_service_pb2.GetTrialRequest(
        name=resources.TrialResource(
            self._owner_id, self._study_id, trial_id
        ).name
    )
    trial = self._service.GetTrial(request)
    return pyvizier.TrialConverter.from_proto(trial)

  def list_trials(self) -> List[pyvizier.Trial]:
    """List all trials."""
    parent = resources.StudyResource(self._owner_id, self._study_id).name
    request = vizier_service_pb2.ListTrialsRequest(parent=parent)
    response = self._service.ListTrials(request)
    return pyvizier.TrialConverter.from_protos(response.trials)

  def list_optimal_trials(self) -> List[pyvizier.Trial]:
    """List only the optimal completed trials."""
    parent = resources.StudyResource(self._owner_id, self._study_id).name
    request = vizier_service_pb2.ListOptimalTrialsRequest(parent=parent)
    response = self._service.ListOptimalTrials(request)
    return pyvizier.TrialConverter.from_protos(response.optimal_trials)

  def list_studies(self) -> List[Dict[str, Any]]:
    """List all studies for the given owner."""
    request = vizier_service_pb2.ListStudiesRequest(
        parent=resources.OwnerResource(self._owner_id).name
    )
    list_studies_response = self._service.ListStudies(request)
    # TODO: Use PyVizier StudyDescriptor instead.
    return [
        json_format.MessageToJson(study)
        for study in list_studies_response.studies
    ]

  def add_trial(self, trial: pyvizier.Trial) -> pyvizier.Trial:
    """Adds a trial.

    Args:
      trial:

    Returns:
      A new trial object from the database. It will have different timestamps,
      newly assigned id, and newly assigned state (either REQUESTED
      or COMPLETED).
    """
    request = vizier_service_pb2.CreateTrialRequest(
        parent=resources.StudyResource(self._owner_id, self._study_id).name,
        trial=pyvizier.TrialConverter.to_proto(trial),
    )
    trial_proto = self._service.CreateTrial(request)
    return pyvizier.TrialConverter.from_proto(trial_proto)

  def delete_trial(self, trial_id: int) -> None:
    """Deletes trial from datastore."""
    request_trial_name = resources.TrialResource(
        self._owner_id, self._study_id, trial_id
    ).name
    request = vizier_service_pb2.DeleteTrialRequest(name=request_trial_name)
    self._service.DeleteTrial(request)
    logging.info('Trial deleted: %s', trial_id)

  def delete_study(self, study_resource_name: Optional[str] = None) -> None:
    """Deletes study from datastore."""
    study_resource_name = study_resource_name or (
        resources.StudyResource(self._owner_id, self._study_id).name
    )
    request = vizier_service_pb2.DeleteStudyRequest(name=study_resource_name)
    self._service.DeleteStudy(request)
    logging.info('Study deleted: %s', study_resource_name)

  def get_study_config(
      self, study_resource_name: Optional[str] = None
  ) -> pyvizier.StudyConfig:
    """Returns the study config."""
    study_resource_name = study_resource_name or (
        resources.StudyResource(self._owner_id, self._study_id).name
    )
    request = vizier_service_pb2.GetStudyRequest(name=study_resource_name)
    study = self._service.GetStudy(request)
    return pyvizier.StudyConfig.from_proto(study.study_spec)

  def get_study_state(
      self, study_resource_name: Optional[str] = None
  ) -> pyvizier.StudyState:
    study_resource_name = study_resource_name or (
        resources.StudyResource(self._owner_id, self._study_id).name
    )
    request = vizier_service_pb2.GetStudyRequest(name=study_resource_name)
    study = self._service.GetStudy(request)
    return pyvizier.StudyStateConverter.from_proto(study.state)

  def set_study_state(
      self,
      state: pyvizier.StudyState,
      study_resource_name: Optional[str] = None,
  ) -> None:
    """Sets the study to a given state."""
    study_resource_name = study_resource_name or (
        resources.StudyResource(self._owner_id, self._study_id).name
    )
    proto_state = pyvizier.StudyStateConverter.to_proto(state)
    request = vizier_service_pb2.SetStudyStateRequest(
        parent=study_resource_name, state=proto_state
    )
    self._service.SetStudyState(request)
    logging.info(
        'Study with resource name %s set to %s state.',
        study_resource_name,
        state,
    )

  def update_metadata(
      self,
      delta: pyvizier.MetadataDelta,
      study_resource_name: Optional[str] = None,
  ) -> None:
    """Updates metadata.

    Args:
      delta: Batch updates to Metadata, consisting of numerous (namespace, key,
        value) objects.
      study_resource_name: An identifier of the study. The full study name will
        be `owners/{owner_id}/studies/{study_id}`.

    Returns:
      None.

    Raises:
      RuntimeError: If service reported an error or if a value could not be
      pickled.
    """
    study_resource_name = study_resource_name or (
        resources.StudyResource(self._owner_id, self._study_id).name
    )
    request = pyvizier.metadata_util.to_request_proto(
        study_resource_name, delta
    )
    response = self._service.UpdateMetadata(request)

    if response.error_details:
      raise RuntimeError(response.error_details)


def create_or_load_study(
    owner_id: str,
    client_id: str,
    study_id: str,
    study_config: pyvizier.StudyConfig,
) -> VizierClient:
  """Factory method for creating or loading a VizierClient.

  This will either create or load the specified study, given
  (owner_id, study_id, study_config). It will create it if it doesn't
  already exist, and load it if someone has already created it.

  Note that once a study is created, you CANNOT modify it with this function.

  This function is designed for use in a distributed system, where many jobs
  initially call create_or_load_study() nearly simultaneously with the same
  `study_config`. In that situation, all clients will end up pointing nicely
  to the same study.

  Args:
      owner_id: An owner id.
      client_id: ID for the VizierClient. See class for notes.
      study_id: Each study is uniquely identified by the tuple (owner_id,
        study_id).
      study_config: Study configuration for Vizier service. If not supplied, it
        will be assumed that the study with the given study_id already exists,
        and will try to retrieve that study.

  Returns:
      A VizierClient object with the specified study created or loaded.

  Raises:
      RuntimeError: Indicates that study_config is supplied but CreateStudy
          failed and GetStudy did not succeed after
          constants.MAX_NUM_TRIES_FOR_STUDIES tries.
      ValueError: Indicates that study_config is not supplied and the study
          with the given study_id does not exist.
  """
  vizier_stub = create_vizier_servicer_or_stub()
  study = study_pb2.Study(
      display_name=study_id, study_spec=study_config.to_proto()
  )
  request = vizier_service_pb2.CreateStudyRequest(
      parent=resources.OwnerResource(owner_id).name, study=study
  )
  # The response study contains a service assigned `name`, and may have been
  # created by this RPC or a previous RPC from another client.
  study = vizier_stub.CreateStudy(request)
  return VizierClient(study.name, client_id, vizier_stub)


def PollingDelay(num_attempts: int, time_scale: float) -> datetime.timedelta:  # pylint:disable=invalid-name
  """Computes a delay to the next attempt to poll the Vizier service.

  This does bounded exponential backoff, starting with $time_scale.
  If $time_scale == 0, it starts with a small time interval, less than
  1 second.

  Args:
    num_attempts: The number of times have we polled and found that the desired
      result was not yet available.
    time_scale: The shortest polling interval, in seconds, or zero. Zero is
      treated as a small interval, less than 1 second.

  Returns:
    A recommended delay interval, in seconds.
  """
  small_interval = 0.3  # Seconds
  interval = max(time_scale, small_interval) * 1.41 ** min(num_attempts, 9)
  return datetime.timedelta(seconds=interval)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""L1 categorical experimenter.

The experimenter's evaluation function counts the number of different parameters
values between the optimal point ('optimum') and the suggestion trial.
"""

import copy
import logging
from typing import Optional, Sequence

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter


class L1CategorialExperimenter(experimenter.Experimenter):
  """Categorical experimenter associated minimum optimization problem."""

  def __init__(
      self,
      *,
      num_categories: Sequence[int],
      optimum: Optional[Sequence[int]] = None,
      seed: Optional[int] = None,
      verbose: bool = False,
  ):
    """Constructor.

    If optimum is not provided, a randomly generated optimum point will be
    created.

    Arguments:
      num_categories: The number of categories in each parameter.
      optimum: Optional list of indices indicating the optimum point. If not
        set, randomly created from seed.
      seed: Optional random generator seed.
      verbose: Whether to show logs.
    """
    rng = np.random.default_rng(seed=seed)
    self._problem = vz.ProblemStatement()
    self._optimum = {}
    for i, num_category in enumerate(num_categories):
      categories = [str(x) for x in range(num_category)]
      param_name = f'c{i}'
      self._problem.search_space.root.add_categorical_param(
          param_name, categories)
      if optimum is None:
        # Define the optimum point by randomally selecting categorical values.
        self._optimum[param_name] = str(rng.integers(low=0, high=num_category))
      elif optimum[i] >= num_category:
        raise ValueError("Optimum point doesn't match category dimensions!")
      else:
        self._optimum[param_name] = str(optimum[i])

    self._problem.metric_information.append(
        vz.MetricInformation(
            name='objective', goal=vz.ObjectiveMetricGoal.MINIMIZE))
    if verbose:
      logging.info('L1CategoricalExperimenter optimum point: %s', self._optimum)

  def evaluate(self, suggestions: Sequence[vz.Trial]):
    for suggestion in suggestions:
      loss = 0.0
      for param_config in self._problem.search_space.parameters:
        if suggestion.parameters[param_config.name].value != self._optimum[
            param_config.name]:
          loss += 1.0
      suggestion.complete(vz.Measurement(metrics={'objective': loss}))

  @property
  def optimal_trial(self) -> vz.Trial:
    """Evaluates and returns the optimal trial."""
    optimal_trial = vz.Trial(parameters=self._optimum)
    self.evaluate([optimal_trial])
    return optimal_trial

  def problem_statement(self):
    """See superclass."""
    return copy.deepcopy(self._problem)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier._src.service import constants
from vizier._src.service import resources
from vizier._src.service import service_policy_supporter
from vizier._src.service import study_pb2
from vizier._src.service import vizier_service
from vizier._src.service.testing import util as test_util
from vizier.service import pyvizier

from absl.testing import absltest


class PythiaSupporterTest(absltest.TestCase):

  def setUp(self):
    self.owner_id = 'my_username'
    self.study_id = '1231223'
    self.study_name = resources.StudyResource(
        owner_id=self.owner_id, study_id=self.study_id
    ).name
    self.vs = vizier_service.VizierServicer(
        database_url=constants.SQL_MEMORY_URL
    )
    self.example_study = test_util.generate_study(self.owner_id, self.study_id)
    self.vs.datastore.create_study(self.example_study)

    self.active_trials = test_util.generate_trials(
        range(1, 7), self.owner_id, self.study_id
    )
    self.succeeded_trial = test_util.generate_trials(
        [7],
        self.owner_id,
        self.study_id,
        state=study_pb2.Trial.State.SUCCEEDED,
        final_measurement=study_pb2.Measurement(),
    )[0]

    for trial in self.active_trials + [self.succeeded_trial]:
      self.vs.datastore.create_trial(trial)

    self.policy_supporter = service_policy_supporter.ServicePolicySupporter(
        self.study_name, self.vs
    )

    super().setUp()

  def test_trial_names_filter(self):
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, trial_ids=[3, 4]
    )

    self.assertEqual(
        trials[0], pyvizier.TrialConverter.from_proto(self.active_trials[2])
    )
    self.assertEqual(
        trials[1], pyvizier.TrialConverter.from_proto(self.active_trials[3])
    )

  def test_min_max_filter(self):
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, min_trial_id=3, max_trial_id=4
    )

    self.assertEqual(
        trials[0], pyvizier.TrialConverter.from_proto(self.active_trials[2])
    )
    self.assertEqual(
        trials[1], pyvizier.TrialConverter.from_proto(self.active_trials[3])
    )

  def test_status_match_filter(self):
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, status_matches=pyvizier.TrialStatus.ACTIVE
    )

    self.assertLen(trials, 6)
    self.assertEqual(
        trials[0], pyvizier.TrialConverter.from_proto(self.active_trials[0])
    )

  def test_raise_value_error(self):
    def should_raise_value_error_fn():
      self.policy_supporter.GetTrials(
          study_guid=self.study_name,
          trial_ids=[1],
          status_matches=pyvizier.TrialStatus.ACTIVE,
      )

      with self.assertRaises(ValueError):
        should_raise_value_error_fn()

  def test_get_study_config(self):
    pythia_problem = self.policy_supporter.GetStudyConfig(self.study_name)
    correct_pythia_problem = pyvizier.StudyConfig.from_proto(
        self.example_study.study_spec
    ).to_problem()
    self.assertEqual(pythia_problem, correct_pythia_problem)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Sparse experimenter.

The SparseExperimenter class wraps another experimenter and expands its search
space with placeholder parameters that don't impact the evaluation.

This experimenter allows testing that a designer/policy is able to optimize an
objective function when only a subset of its parameters affect the function
values.
"""

import copy
from typing import Optional, Sequence

import attr
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter


# These default values are used for defining the placeholder parameters.
_DEFAULT_FLOAT_BOUNDS = (-5.0, 5.0)
_DEFAULT_INT_BOUNDS = (-5, 5)

_DEFAULT_DISCRETE_FEASIBLE_VALUES = (0, 1, 2, 3, 4)
_DEFAULT_CATEGORICAL_FEASIBLE_VALUES = ('a', 'b', 'c', 'd', 'e', 'f')


class SparseExperimenter(experimenter.Experimenter):
  """Sparse experimenter.

  The sparse experimenter allows expanding an experimenter's search space with
  additional placeholder parameters that don't have affect during evaluation.
  """

  def __init__(
      self,
      experiment: experimenter.Experimenter,
      search_space: vz.SearchSpace,
      prefix: str = '_SPARSE',
  ):
    """Initializes a sparse experimenter.

    The sparse experimenter is constructed by adding a copy of parameters from
    $search_space to the experimenter search space. Newly added parameter names
    are prefixed with $prefix.

    Arguments:
      experiment: An experimenter to add sparse parameters to.
      search_space: A search space to use for the sparse parameters.
      prefix: The sparse parameter name prefix (shouldn't already exist in the
        search space of 'experiment').
    """
    super().__init__()
    self._sparse_param_prefix = prefix
    self._experimenter = experiment
    self._search_space = copy.deepcopy(search_space)
    self._prefix = prefix

    problem = experiment.problem_statement()
    for pc in search_space.parameters:
      # Add a copy of the parameter config with a modified name.
      problem.search_space.add(attr.evolve(pc, name=prefix + '_' + pc.name))
    self._problem_statement = problem

  def evaluate(self, suggestions: Sequence[vz.Trial]):
    """Evaluates and completes trials on sparse subset parameters."""
    # Evaluates trials on non-sparse parameters only determined by the prefix.
    original_params = []
    for trial in suggestions:
      original_params.append(trial.parameters)
      trial.parameters = {
          param_name: param_value
          for param_name, param_value in trial.parameters.items()
          if not param_name.startswith(self._sparse_param_prefix)
      }
    self._experimenter.evaluate(suggestions)
    # Restores the original parameters with the sparse parameters.
    for trial, params in zip(suggestions, original_params):
      trial.parameters = params

  def problem_statement(self) -> vz.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  @classmethod
  def create(
      cls,
      experiment: experimenter.Experimenter,
      float_count: int,
      int_count: int,
      discrete_count: int,
      categorical_count: int,
      float_min_value: Optional[float] = None,
      float_max_value: Optional[float] = None,
      int_min_value: Optional[int] = None,
      int_max_value: Optional[int] = None,
      discrete_feasible_values: Optional[list[int]] = None,
      categorical_feasible_values: Optional[list[str]] = None,
  ) -> 'SparseExperimenter':
    """Creates a sparse experimenter with different parameter types.

    Arguments:
      experiment: The experimenter to add sparsity to.
      float_count: The number of FLOAT sparse parameters to add.
      int_count: The number of INT sparse parameters to add.
      discrete_count: The number of DISCRETE sparse parameters to add.
      categorical_count: The number of CATEGORICAL sparse parameters to add.
      float_min_value: The FLOAT sparse parameter min value.
      float_max_value: The INTEGER sparse parameter max value.
      int_min_value: The FLOAT sparse parameter min value.
      int_max_value: The INTEGER sparse parameter max value.
      discrete_feasible_values: The DISCRETE sparse parameter feasible values.
      categorical_feasible_values: The CTEGORICAL sparse parameter feasible
        values.

    Returns:
      The sparse experimenter.
    """
    sparse_search_space = vz.SearchSpace()
    for idx in range(float_count):
      sparse_search_space.root.add_float_param(
          name='FLOAT' + str(idx),
          min_value=float_min_value or _DEFAULT_FLOAT_BOUNDS[0],
          max_value=float_max_value or _DEFAULT_FLOAT_BOUNDS[1],
      )
    for idx in range(int_count):
      sparse_search_space.root.add_int_param(
          name='INT' + str(idx),
          min_value=float_min_value or _DEFAULT_INT_BOUNDS[0],
          max_value=float_max_value or _DEFAULT_INT_BOUNDS[1],
      )
    for idx in range(discrete_count):
      sparse_search_space.root.add_discrete_param(
          name='DISCRETE' + str(idx),
          feasible_values=categorical_feasible_values
          or _DEFAULT_DISCRETE_FEASIBLE_VALUES,
      )
    for idx in range(categorical_count):
      sparse_search_space.root.add_categorical_param(
          name='CATEGORICAL' + str(idx),
          feasible_values=categorical_feasible_values
          or _DEFAULT_CATEGORICAL_FEASIBLE_VALUES,
      )
    return cls(experiment, sparse_search_space)
